"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const math = require("mathjs");
const dymo_core_1 = require("dymo-core");
var Features;
(function (Features) {
    Features[Features["TempoA"] = 0] = "TempoA";
    Features[Features["TempoB"] = 1] = "TempoB";
    Features[Features["TempoRatio"] = 2] = "TempoRatio";
    Features[Features["TempoMultiplicity"] = 3] = "TempoMultiplicity";
    Features[Features["RegularityA"] = 4] = "RegularityA";
    Features[Features["RegularityB"] = 5] = "RegularityB";
    Features[Features["RegularityProduct"] = 6] = "RegularityProduct";
    Features[Features["KeyA"] = 7] = "KeyA";
    Features[Features["KeyB"] = 8] = "KeyB";
    Features[Features["KeyDistance"] = 9] = "KeyDistance";
})(Features = exports.Features || (exports.Features = {}));
/* {[0]: 0, [7,5]: 1, [2,10]: 2, [9,3]: 3, [4,8]: 4, [11,1]: 5, [6]: 6} */
const TONAL_DISTANCES = {
    0: 0,
    1: 5,
    2: 2,
    3: 3,
    4: 4,
    5: 1,
    6: 6,
    7: 1,
    8: 4,
    9: 3,
    10: 2,
    11: 5
};
class Analyzer {
    constructor(store) {
        this.store = store;
        this.beatsCache = new Map();
        this.keysCache = new Map();
        this.tempoCache = new Map();
    }
    getAllFeatures(track1, track2) {
        return __awaiter(this, void 0, void 0, function* () {
            return [
                yield this.getTempo(track1),
                yield this.getTempo(track2),
                yield this.getTempoRatio(track1, track2),
                yield this.getTempoMultiple(track1, track2),
                yield this.getRegularity(track1),
                yield this.getRegularity(track2),
                yield this.getRegularityProduct(track1, track2),
                yield this.getKey(track1),
                yield this.getKey(track2),
                yield this.getKeyDistance(track1, track2)
            ];
            /*USED FOR EARLY EXPERIMENTS
              return [
              await this.getTempo(track1),
              await this.getTempo(track2),
              await this.getTempoRatio(track1, track2),
              await this.getTempoRatio(track2, track1),
              await this.getTempoMultiple(track1, track2),
              await this.getTempoMultiple(track2, track1),
              await this.getRegularity(track1),
              await this.getRegularity(track2), //TODO ADD REGULARITY PRODUCT OR RATIO
              await this.getKey(track1),
              await this.getKey(track2),
              await this.getKeyDistance(track1, track2)
            ]*/
        });
    }
    findCuePoint(trackUri) {
        return __awaiter(this, void 0, void 0, function* () {
            const loudnesses = yield this.getBarLoudnesses(trackUri);
            const indexOfLastIncrease = loudnesses.findIndex((l, i) => loudnesses[i + 1] < l);
            const initialEighth = loudnesses.slice(0, loudnesses.length / 8);
            const indexOfInitialMax = loudnesses.indexOf(_.max(initialEighth));
            const localMaxSquared = this.findLocalMaxes(this.findLocalMaxes(loudnesses));
            const indexOfLocalMaxSquared = loudnesses.indexOf(localMaxSquared[0]);
            console.log(indexOfLastIncrease, indexOfInitialMax, indexOfLocalMaxSquared);
            return indexOfLocalMaxSquared;
        });
    }
    findLocalMaxes(values) {
        return values.filter((v, i) => v > values[i - 1] && v > values[i + 1]);
    }
    getKeyDistance(track1Uri, track2Uri) {
        return __awaiter(this, void 0, void 0, function* () {
            let dist = Math.abs((yield this.getKey(track1Uri)) - (yield this.getKey(track2Uri)));
            return TONAL_DISTANCES[dist];
        });
    }
    getKey(trackUri) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.keysCache.has(trackUri)) {
                const key = yield this.store.findFeatureValue(trackUri, dymo_core_1.uris.CONTEXT_URI + "key");
                if (key) {
                    this.keysCache.set(trackUri, key.length ? key[0] : key);
                }
            }
            return this.keysCache.get(trackUri);
        });
    }
    getTempo(trackUri) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.tempoCache.has(trackUri)) {
                const durations = yield this.getBeatDurations(trackUri);
                if (durations && durations.length) {
                    this.tempoCache.set(trackUri, 60 / math.mean(durations));
                }
            }
            return this.tempoCache.get(trackUri);
        });
    }
    //a symmetrical ratio < 1
    getTempoMultiple(track1Uri, track2Uri) {
        return __awaiter(this, void 0, void 0, function* () {
            let tempoRatio = (yield this.getTempo(track1Uri)) / (yield this.getTempo(track2Uri));
            //make it larger than 1
            tempoRatio = tempoRatio < 1 ? 1 / tempoRatio : tempoRatio;
            let tempoMultiple = tempoRatio % 1;
            //back to [0,1]
            return tempoMultiple > 1 ? 1 / tempoMultiple : tempoMultiple;
        });
    }
    //a symmetrical ratio < 1
    getTempoRatio(track1Uri, track2Uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const tempoRatio = (yield this.getTempo(track1Uri)) / (yield this.getTempo(track2Uri));
            return tempoRatio > 1 ? 1 / tempoRatio : tempoRatio;
        });
    }
    hasRegularBeats(trackUri) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.getRegularity(trackUri)) < .1;
        });
    }
    getRegularityProduct(track1Uri, track2Uri) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.getRegularity(track1Uri)) * (yield this.getRegularity(track2Uri));
        });
    }
    getRegularity(trackUri) {
        return __awaiter(this, void 0, void 0, function* () {
            const durations = yield this.getBeatDurations(trackUri);
            if (durations && durations.length) {
                return math.std(durations);
            }
        });
    }
    tempoSimilar(track1Uri, track2Uri) {
        return __awaiter(this, void 0, void 0, function* () {
            const ratio = yield this.getTempoRatio(track1Uri, track2Uri);
            return this.isSimilar(1, ratio);
        });
    }
    isSimilar(n1, n2) {
        //TODO MAKE POWER-BASED DISTANCE
        return Math.abs(n1 - n2) < .1;
    }
    getBarLoudnesses(trackUri) {
        return __awaiter(this, void 0, void 0, function* () {
            const bars = yield this.store.findParts(trackUri);
            return yield this.findFeatureSeries(bars, dymo_core_1.uris.CONTEXT_URI + "loudness");
        });
    }
    getBeatDurations(trackUri) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.beatsCache.has(trackUri)) {
                const bars = yield this.store.findParts(trackUri);
                const beats = _.flatten(yield Promise.all(bars.map(p => this.store.findParts(p))));
                const durations = yield this.findFeatureSeries(beats, dymo_core_1.uris.DURATION_FEATURE);
                this.beatsCache.set(trackUri, durations);
            }
            return this.beatsCache.get(trackUri);
        });
    }
    findFeatureSeries(dymos, featureUri) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield Promise.all(dymos.map(b => this.findFeature(b, featureUri)));
        });
    }
    findFeature(dymo, featureUri) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.store.findFeatureValue(dymo, featureUri));
        });
    }
}
exports.Analyzer = Analyzer;
