"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const _ = require("lodash");
const dymo_player_1 = require("dymo-player");
const dymo_core_1 = require("dymo-core");
const mix_generator_1 = require("./mix-generator");
const types_1 = require("./types");
const analyzer_1 = require("./analyzer");
const feature_extractor_1 = require("./feature-extractor");
const decision_tree_1 = require("./decision-tree");
const standard_tree_1 = require("./standard-tree");
class AutoDj {
    constructor(featureService, decisionType, decisionTree = standard_tree_1.STANDARD_TREE, defaultTransitionType = types_1.TransitionType.Beatmatch) {
        this.featureService = featureService;
        this.decisionType = decisionType;
        this.defaultTransitionType = defaultTransitionType;
        this.previousPlayingDymos = [];
        this.beatsPlayed = 0;
        this.previousTracks = [];
        this.scheduledTransitions = [];
        this.transitionsObservable = new rxjs_1.Subject();
        this.transitionsObserved = 0;
        this.ready = this.init(decisionTree);
    }
    init(decisionTree) {
        return __awaiter(this, void 0, void 0, function* () {
            this.decisionTree = new decision_tree_1.DecisionTree(decisionTree);
            this.player = new dymo_player_1.DymoPlayer({
                useWorkers: true,
                scheduleAheadTime: 0.5,
                loadAheadTime: 1,
                useTone: true
            });
            yield this.player.init('https://raw.githubusercontent.com/dynamic-music/dymo-core/master/ontologies/'); //'https://dynamic-music.github.io/dymo-core/ontologies/')
            this.store = this.player.getDymoManager().getStore();
            this.dymoGen = new dymo_core_1.DymoGenerator(false, this.store);
            this.mixGen = new mix_generator_1.MixGenerator(this.dymoGen, this.player, this.updateTransitionObservable.bind(this));
            this.analyzer = new analyzer_1.Analyzer(this.store);
            if (!this.featureService) {
                this.featureService = new feature_extractor_1.FeatureExtractor(yield this.player.getAudioBank());
            }
        });
    }
    isReady() {
        return this.ready;
    }
    getBeatObservable() {
        return (this.player.getPlayingDymoUris())
            .pipe(operators_1.filter(playingDymos => {
            //simple way to check wether there are new dymos playing, thus a new beat
            const nChanged = _.difference(playingDymos, this.previousPlayingDymos).length;
            this.previousPlayingDymos = playingDymos;
            return nChanged > 0;
        }))
            .pipe(operators_1.map(() => this.beatsPlayed++));
    }
    getTransitionObservable() {
        return this.transitionsObservable.asObservable();
    }
    updateTransitionObservable() {
        this.transitionsObservable.next(this.scheduledTransitions[this.transitionsObserved]);
        this.transitionsObserved++;
    }
    transitionToTrack(audioUri) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            yield this.resetIfStopped();
            //make sure audio is loaded before attempting transition
            yield this.player.getAudioBank().preloadBuffers([audioUri]);
            const newTrack = yield this.extractFeaturesAndAddDymo(audioUri);
            return this.internalTransition(audioUri, { trackUri: newTrack });
        });
    }
    playDjSet(audioUris, numBars, autoCue) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.ready;
            this.mapSeries(audioUris, (a, i) => __awaiter(this, void 0, void 0, function* () {
                console.log("buffering", a);
                yield this.player.getAudioBank().preloadBuffers([a]);
                console.log("preloaded");
                yield this.addTrackToMix(a, -1, numBars, autoCue, 4);
                console.log("added, length now", (yield this.store.findParts(this.mixGen.getMixDymo())).length);
            }));
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = this.mixGen.getMixDymo();
            this.reset();
            yield this.player.stopUri(uri);
        });
    }
    addTrackToMix(audioUri, position, numBars, autoCue, duration) {
        return __awaiter(this, void 0, void 0, function* () {
            const newTrack = yield this.extractFeaturesAndAddDymo(audioUri);
            const options = { trackUri: newTrack };
            if (autoCue) {
                options.cueOffset = yield this.analyzer.findCuePoint(newTrack);
            }
            options.numBars = numBars;
            options.duration = duration;
            options.position = position;
            yield this.internalTransition(audioUri, options);
        });
    }
    internalTransition(audioUri, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const features = yield this.getTransitionFeatures(options.trackUri);
            const transition = yield this.transitionBasedOnDecisionType(options, features);
            transition.features = features;
            this.previousTracks.push(options.trackUri);
            this.scheduledTransitions.push(transition);
            transition.names = [audioUri];
            return transition;
        });
    }
    resetIfStopped() {
        if (this.previousTracks.length > 0 && !this.player.isPlaying(this.mixGen.getMixDymo())) {
            return this.reset();
        }
    }
    reset() {
        this.previousTracks = [];
        return this.mixGen.init();
    }
    getTransitionFeatures(newTrack) {
        return __awaiter(this, void 0, void 0, function* () {
            //await this.analyzer.findCuePoint(newTrack);
            if (this.previousTracks.length > 0) {
                const oldTrack = _.last(this.previousTracks);
                return this.analyzer.getAllFeatures(oldTrack, newTrack);
            }
        });
    }
    extractFeaturesAndAddDymo(audioUri) {
        return __awaiter(this, void 0, void 0, function* () {
            let beats = yield this.featureService.getBeats(audioUri);
            //drop initial and final incomplete bars
            beats = _.dropWhile(beats, b => b.label.value !== "1");
            beats = _.dropRightWhile(beats, b => b.label.value !== "4");
            const segments = beats.map(b => ({ time: b.time.value, value: b.label.value }));
            const newTrack = yield dymo_core_1.DymoTemplates.createAnnotatedBarAndBeatDymo2(this.dymoGen, audioUri, segments);
            const keys = yield this.featureService.getKeys(audioUri);
            if (keys) {
                yield this.addFeature("key", keys, newTrack, dymo_core_1.SUMMARY.MODE);
            }
            const loudnesses = yield this.featureService.getLoudnesses(audioUri);
            if (loudnesses) {
                yield this.addFeature("loudness", loudnesses, newTrack, dymo_core_1.SUMMARY.MEAN);
            }
            yield this.player.getDymoManager().loadFromStore(newTrack);
            return newTrack;
        });
    }
    addFeature(name, values, dymoUri, summaryMode) {
        return __awaiter(this, void 0, void 0, function* () {
            if (values) {
                this.dymoGen.setSummarizingMode(summaryMode);
                const data = values.map(f => ({ time: f.time.value, value: f.value }));
                yield this.dymoGen.addFeature(name, data, dymoUri);
            }
        });
    }
    transitionBasedOnDecisionType(options, features) {
        return __awaiter(this, void 0, void 0, function* () {
            let transition;
            if (this.previousTracks.length == 0) {
                transition = yield this.mixGen.startMixWithFadeIn(options);
            }
            else if (this.decisionType == types_1.DecisionType.Default) {
                transition = yield this.mixGen[this.defaultTransitionType](options);
            }
            else if (this.decisionType == types_1.DecisionType.Random) {
                transition = yield this.randomTransition(options);
            }
            else if (this.decisionType == types_1.DecisionType.FiftyFifty) {
                //fiftyfifty random and decision tree
                transition = Math.random() > 0.5 ? yield this.randomTransition(options)
                    : yield this.decisionTreeTransition(options, features);
            }
            else {
                transition = yield this.decisionTreeTransition(options, features);
            }
            if (this.decisionType != types_1.DecisionType.FiftyFifty) {
                transition.decision = this.decisionType;
            }
            this.player.playUri(this.mixGen.getMixDymo());
            return transition;
        });
    }
    randomTransition(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const randomTransition = _.sample(mix_generator_1.AVAILABLE_TRANSITIONS);
            const transition = yield this.mixGen[randomTransition](options);
            transition.decision = types_1.DecisionType.Random;
            return transition;
        });
    }
    decisionTreeTransition(options, features) {
        return __awaiter(this, void 0, void 0, function* () {
            console.log("asking tree");
            const transitionType = this.decisionTree.classify(features);
            const transition = yield this.mixGen[transitionType](options);
            transition.decision = types_1.DecisionType.DecisionTree;
            console.log(transitionType);
            return transition;
        });
    }
    mapSeries(array, func) {
        return __awaiter(this, void 0, void 0, function* () {
            let result = [];
            for (let i = 0; i < array.length; i++) {
                result.push(yield func(array[i], i));
            }
            return result;
        });
    }
}
exports.AutoDj = AutoDj;
