"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const dymo_core_1 = require("dymo-core");
const types_1 = require("./types");
const transition_observer_1 = require("./transition-observer");
exports.AVAILABLE_TRANSITIONS = _.values(types_1.TransitionType);
const TRANSITION_OFFSET = 1; //number of bars from current position an asap transition starts
class MixGenerator {
    constructor(generator, player, transitionObserverFunction) {
        this.generator = generator;
        this.player = player;
        this.transitionObserverFunction = transitionObserverFunction;
        this.transitionConstraints = []; //ARRAYS OF CONSTRAINT URIS FOR NOW
        this.store = generator.getStore();
        this.expressionGen = new dymo_core_1.ExpressionGenerator(this.store);
        this.init();
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            this.tracks = [];
            this.mixDymoUri = yield this.generator.addDymo();
        });
    }
    getMixDymo() {
        return this.mixDymoUri;
    }
    startMixWithFadeIn(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const numBars = options.duration ? options.duration : 2;
            const newTrackBars = yield this.registerTrackAndGetBars(options);
            const [duration, uris] = yield this.applyFadeIn(newTrackBars.slice(0, numBars));
            return this.endTransition(newTrackBars, types_1.TransitionType.FadeIn, duration, uris);
        });
    }
    slam(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = yield this.initTransition(options);
            return this.endTransition(state.newTrackBars, types_1.TransitionType.Slam, 0);
        });
    }
    beatRepeat(options, times = 3) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = yield this.initTransition(options);
            //add reverb to last bar
            let lastBar = yield this.findLastBar();
            let lastBeat = yield this.store.findPartAt(lastBar, 3);
            yield this.store.setParameter(lastBeat, dymo_core_1.uris.DELAY, 0.5);
            //add silence for first part of bar
            let lastBarDuration = yield this.store.findFeatureValue(state.removedOldTrackBars[0], dymo_core_1.uris.DURATION_FEATURE);
            yield this.addSilence(lastBarDuration / 2);
            //beat repeat
            let firstBarBeats = yield this.store.findParts(state.newTrackBars[0]);
            yield this.addPartsToMix(_.fill(Array(times), firstBarBeats[0]));
            return this.endTransition(state.newTrackBars, types_1.TransitionType.BeatRepeat, 2); //duration just an estimate for now
        });
    }
    echoFreeze(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const numBarsBreak = options.duration ? options.duration : 2;
            const state = yield this.initTransition(options);
            //delay out last bar
            let lastBar = yield this.findLastBar();
            yield this.store.setParameter(lastBar, dymo_core_1.uris.DELAY, 1);
            //add silence for n bars
            const lastBarDuration = yield this.store.findFeatureValue(lastBar, dymo_core_1.uris.DURATION_FEATURE);
            const silenceDuration = lastBarDuration * numBarsBreak;
            yield this.addSilence(silenceDuration);
            return this.endTransition(state.newTrackBars, types_1.TransitionType.EchoFreeze, lastBarDuration + silenceDuration);
        });
    }
    effects(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const numBars = options.duration ? options.duration : 2;
            const state = yield this.initTransition(options);
            //add effects to a few bars
            const effectBars = state.removedOldTrackBars.slice(0, numBars);
            const duration = yield this.getTotalDuration(effectBars);
            const effectsRamp = yield this.addRampWithTrigger(duration);
            const reverb = yield this.makeRampConstraint(effectsRamp, effectBars, 'Reverb(d) == r/3');
            yield this.addPartsToMix(effectBars);
            return this.endTransition(state.newTrackBars, types_1.TransitionType.Effects, duration, [effectsRamp, reverb]);
        });
    }
    powerDown(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const numBars = options.duration ? options.duration : 2;
            const numBarsBreak = 0;
            const state = yield this.initTransition(options);
            //add power down to a few bars
            const powerBars = state.removedOldTrackBars.slice(0, numBars);
            //avg duration with linear ramp power down
            const duration = 2 * (yield this.getTotalDuration(powerBars));
            const powerRamp = yield this.addRampWithTrigger(duration);
            const powerDown = yield this.makeRampConstraint(powerRamp, powerBars, 'PlaybackRate(d) == 1-r');
            yield this.addPartsToMix(powerBars);
            //add silence for n bars
            const silenceDuration = (duration / 2 / numBars) * numBarsBreak;
            yield this.addSilence(silenceDuration);
            //add new track
            return this.endTransition(state.newTrackBars, types_1.TransitionType.PowerDown, duration + silenceDuration, [powerRamp, powerDown]);
        });
    }
    crossfade(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const numBars = options.duration ? options.duration : 3;
            const state = yield this.initTransition(options);
            const newTrackTrans = state.newTrackBars.slice(0, numBars);
            const duration = yield this.getTotalDuration(newTrackTrans);
            const oldTrackTrans = yield this.getInitialBars(state.removedOldTrackBars, duration);
            let uris = yield this.applyCrossfade(oldTrackTrans, newTrackTrans, duration);
            yield this.addAligned(oldTrackTrans, newTrackTrans);
            return this.endTransition(state.newTrackBars.slice(numBars), types_1.TransitionType.Crossfade, duration, uris);
        });
    }
    beatmatchCrossfade(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const numBars = options.duration ? options.duration : 3;
            const state = yield this.initTransition(options);
            const newTrackTrans = state.newTrackBars.slice(0, numBars);
            const oldTrackTrans = state.removedOldTrackBars.slice(0, numBars);
            const duration = (yield this.getTotalDuration(oldTrackTrans.concat(newTrackTrans))) / 2 - 0.5; //minus schedule ahead time for more tempo smoothness!
            //add constraints, controls, and triggers
            let uris = yield this.applyCrossfade(oldTrackTrans, newTrackTrans, duration);
            //only beatmatch if same number of bars
            if (newTrackTrans.length == oldTrackTrans.length) {
                uris = uris.concat(yield this.applyBeatmatch(oldTrackTrans, newTrackTrans, uris[0]));
            }
            //add transition part
            yield this.addZipped(oldTrackTrans, newTrackTrans);
            return this.endTransition(state.newTrackBars.slice(numBars), types_1.TransitionType.Beatmatch, duration, uris);
        });
    }
    addPartsToMix(parts) {
        return __awaiter(this, void 0, void 0, function* () {
            return Promise.all(parts.map(p => this.store.addPart(this.mixDymoUri, p)));
        });
    }
    addSilence(duration) {
        return __awaiter(this, void 0, void 0, function* () {
            if (duration > 0) {
                let silence = yield this.generator.addDymo();
                yield this.store.setFeature(silence, dymo_core_1.uris.DURATION_FEATURE, duration);
                yield this.addPartsToMix([silence]);
            }
        });
    }
    addAligned(bars1, bars2) {
        return __awaiter(this, void 0, void 0, function* () {
            let bars1Seq = yield this.generator.addDymo(null, null, dymo_core_1.uris.SEQUENCE);
            yield Promise.all(bars1.map(p => this.store.addPart(bars1Seq, p)));
            let bars2Seq = yield this.generator.addDymo(null, null, dymo_core_1.uris.SEQUENCE);
            yield Promise.all(bars2.map(p => this.store.addPart(bars2Seq, p)));
            return yield this.generator.addConjunction(this.mixDymoUri, [bars1Seq, bars2Seq]);
        });
    }
    addZipped(bars1, bars2) {
        return __awaiter(this, void 0, void 0, function* () {
            Promise.all(_.zip(bars1, bars2).map(bp => this.generator.addConjunction(this.mixDymoUri, bp)));
        });
    }
    applyBeatmatch(oldTrackBars, newTrackBars, rampUri) {
        return __awaiter(this, void 0, void 0, function* () {
            //create tempo transition
            let tempoParam = yield this.generator.addCustomParameter(dymo_core_1.uris.CONTEXT_URI + "Tempo");
            let newTempo = yield this.getTempoFromBars(newTrackBars);
            let oldTempo = yield this.getTempoFromBars(oldTrackBars);
            let tempoTransition = yield this.makeSetsConstraint([['t', [tempoParam]], ['r', [rampUri]]], 't == r*' + newTempo + '+(1-r)*' + oldTempo);
            //create beatmatch
            let beats = _.flatten(yield Promise.all(oldTrackBars.concat(newTrackBars).map(p => this.store.findParts(p))));
            let beatMatch = yield this.makeSetsConstraint([['d', beats], ['t', [tempoParam]]], 'TimeStretchRatio(d) == t/60*DurationFeature(d)');
            console.log("beatmatched between tempos", oldTempo, newTempo);
            return [tempoTransition, beatMatch];
        });
    }
    applyFadeIn(newTrackBarsParts) {
        return __awaiter(this, void 0, void 0, function* () {
            let fadeDuration = yield this.getTotalDuration(newTrackBarsParts);
            let fadeRamp = yield this.addRampWithTrigger(fadeDuration);
            let fadeIn = yield this.makeRampConstraint(fadeRamp, newTrackBarsParts, 'Amplitude(d) == r');
            console.log("fading in for", newTrackBarsParts.length, "bars (" + fadeDuration + " seconds)");
            return [fadeDuration, [fadeRamp, fadeIn]];
        });
    }
    applyCrossfade(oldTrackParts, newTrackParts, duration) {
        return __awaiter(this, void 0, void 0, function* () {
            let fadeRamp = yield this.addRampWithTrigger(duration);
            let fadeIn = yield this.makeRampConstraint(fadeRamp, newTrackParts, 'Amplitude(d) == r');
            let fadeOut = yield this.makeRampConstraint(fadeRamp, oldTrackParts, 'Amplitude(d) == 1-r');
            console.log("crossfading for", newTrackParts.length, "bars (" + duration + " seconds)");
            return [fadeRamp, fadeIn, fadeOut];
        });
    }
    //TODO dymo-core throws the occasional error due to list editing concurrency problem
    addRandomBeatToLoop(trackUri, loopDuration = 2) {
        return __awaiter(this, void 0, void 0, function* () {
            let currentBeats = yield this.store.findParts(this.mixDymoUri);
            //find a random beat in the track
            let bars = yield this.registerTrackAndGetBars({ trackUri: trackUri });
            let randomBar = bars[_.random(bars.length)];
            let randomBeat = (yield this.store.findParts(randomBar))[_.random(4)];
            if (currentBeats.length == 0) {
                //add silence at beginning and end of loop to ensure constant length :/
                let silenceUri = yield this.generator.addDymo(this.mixDymoUri);
                yield this.store.setParameter(silenceUri, dymo_core_1.uris.ONSET, 0);
                silenceUri = yield this.generator.addDymo(this.mixDymoUri);
                yield this.store.setParameter(silenceUri, dymo_core_1.uris.ONSET, loopDuration);
                currentBeats = yield this.store.findParts(this.mixDymoUri);
            }
            //set a random onset and add the beat to the loop at correct position
            let currentOnsets = yield Promise.all(currentBeats.map(b => this.store.findParameterValue(b, dymo_core_1.uris.ONSET)));
            let randomOnset = _.random(loopDuration, true);
            yield this.store.setParameter(randomBeat, dymo_core_1.uris.ONSET, randomOnset);
            let beatPosition = currentOnsets.filter(o => o < randomOnset).length;
            return this.store.insertPartAt(this.mixDymoUri, randomBeat, beatPosition);
        });
    }
    transitionImmediatelyToRandomBars(trackUri, numBars = 2) {
        return __awaiter(this, void 0, void 0, function* () {
            let bars = yield this.registerTrackAndGetBars({ trackUri: trackUri });
            let randomBar = _.random(bars.length - numBars);
            return Promise.all(bars.slice(randomBar, randomBar + numBars).map(p => this.store.addPart(this.mixDymoUri, p)));
        });
    }
    /**removes old track until current position + offset, registers new track and gets bars*/
    initTransition(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const newTrackBars = yield this.registerTrackAndGetBars(options);
            const length = (yield this.store.findParts(this.mixDymoUri)).length;
            let position;
            if (!options.position) {
                const currentPos = yield this.player.getPosition(this.mixDymoUri);
                position = currentPos + TRANSITION_OFFSET;
            }
            else if (options.position < 0) {
                position = length - options.duration;
            }
            else {
                position = options.position;
            }
            const oldTrackBars = position < length ?
                yield this.store.removeParts(this.mixDymoUri, position) : [];
            return { removedOldTrackBars: oldTrackBars, newTrackBars: newTrackBars };
        });
    }
    /**adds new track bars, loads constraints and controls, and returns transition object*/
    endTransition(newTrackBars, type, duration, transitionUris) {
        return __awaiter(this, void 0, void 0, function* () {
            if (transitionUris) {
                const loaded = yield this.player.getDymoManager().loadFromStore(...transitionUris);
                //TODO NOW ADD CONSTRAINT TRIGGERS (ACTUALLY)
                yield this.addConstraintTriggers(loaded.constraintUris);
                this.transitionConstraints.push(loaded.constraintUris);
            }
            yield this.addPartsToMix(newTrackBars);
            return this.getTransitionObject(type, duration);
        });
    }
    registerTrackAndGetBars(options) {
        return __awaiter(this, void 0, void 0, function* () {
            this.tracks.push(options.trackUri);
            let bars = yield this.store.findParts(options.trackUri);
            const offset = options.cueOffset ? options.cueOffset : 0;
            const numBars = options.numBars ? offset + options.numBars : undefined;
            return bars.slice(offset, numBars);
        });
    }
    addRampWithTrigger(duration) {
        return __awaiter(this, void 0, void 0, function* () {
            const rampUri = yield this.generator.addRampControl(0, duration, 100);
            yield this.addControlTrigger(rampUri);
            return rampUri;
        });
    }
    /** adds a trigger to init the the transition, calls an optional function when
      triggered */
    addControlTrigger(controlUri) {
        return __awaiter(this, void 0, void 0, function* () {
            const trigger = yield this.store
                .setControlParam(controlUri, dymo_core_1.uris.AUTO_CONTROL_TRIGGER, 0);
            new transition_observer_1.TransitionObserver(this.store, trigger, this.transitionObserverFunction);
            yield this.generator.addEvent(this.mixDymoUri, trigger, 1);
        });
    }
    addConstraintTriggers(newUris) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.transitionConstraints.length > 1) {
                const previousConstraints = _.last(this.transitionConstraints);
                /*TODO ADD EVENT TO DEACTIVATE PREVIOUS CONSTRAINTS AND ACTIVATE NEW ONES
                this.store.deactivateConstraints(previousConstraints);*/
            }
        });
    }
    makeCrossfade(rampUri, oldTrackUris, newTrackUris) {
        return __awaiter(this, void 0, void 0, function* () {
            var fadeOut = yield this.makeRampConstraint(rampUri, oldTrackUris, 'Amplitude(d) == 1-r');
            var fadeIn = yield this.makeRampConstraint(rampUri, newTrackUris, 'Amplitude(d) == r');
            /*var fadeOut2 = await this.makeRampConstraint(rampUri, oldTrackUris, 'DurationRatio(d) == 1/(1-r)');
            var fadeIn2 = await this.makeRampConstraint(rampUri, newTrackUris, 'DurationRatio(d) == 1/r');*/
            return [fadeOut, fadeIn].filter(c => c); //remove undefined
        });
    }
    makeRampConstraint(rampUri, dymoUris, expression) {
        if (dymoUris.length > 0) {
            return this.makeSetsConstraint([['d', dymoUris], ['r', [rampUri]]], expression);
        }
    }
    makeSetsConstraint(sets, expression) {
        let vars = sets.map(s => '∀ ' + s[0] + ' in ' + JSON.stringify(s[1]) + ' => ').join('');
        return this.expressionGen.addConstraint(this.mixDymoUri, vars + expression, true);
    }
    getTempoFromBars(barUris) {
        return __awaiter(this, void 0, void 0, function* () {
            let avgDuration = _.mean(yield this.getFeature(barUris, dymo_core_1.uris.DURATION_FEATURE));
            return 60 / (avgDuration / 4);
        });
    }
    //returns an initial segment of bars with at most the given duration
    getInitialBars(bars, duration) {
        return __awaiter(this, void 0, void 0, function* () {
            let currentDuration = 0;
            return _.takeWhile(bars, (b, i) => __awaiter(this, void 0, void 0, function* () {
                currentDuration += yield this.store.findFeatureValue(b, dymo_core_1.uris.DURATION_FEATURE);
                return currentDuration < duration;
            }));
        });
    }
    /**returns the last bars*/
    findLastBar() {
        return __awaiter(this, void 0, void 0, function* () {
            //return (await this.store.findParts(this.mixDymoUri)).slice(-n);
            return _.last(yield this.store.findParts(this.mixDymoUri));
        });
    }
    getTotalDuration(dymoUris) {
        return __awaiter(this, void 0, void 0, function* () {
            return _.sum(yield this.getFeature(dymoUris, dymo_core_1.uris.DURATION_FEATURE));
        });
    }
    getFeature(dymoUris, featureUri) {
        return __awaiter(this, void 0, void 0, function* () {
            return Promise.all(dymoUris.map(d => this.store.findFeatureValue(d, featureUri)));
        });
    }
    getTransitionObject(type, duration) {
        return {
            date: new Date(Date.now()),
            user: null,
            rating: null,
            names: null,
            features: null,
            decision: null,
            type: type,
            parameters: null,
            duration: duration
        };
    }
}
exports.MixGenerator = MixGenerator;
